"use strict";

var _interopRequireDefault = require("@babel/runtime-corejs2/helpers/interopRequireDefault");

var _Object$defineProperty = require("@babel/runtime-corejs2/core-js/object/define-property");

_Object$defineProperty(exports, "__esModule", {
  value: true
});

exports.Gemstone = void 0;

var _stringify = _interopRequireDefault(require("@babel/runtime-corejs2/core-js/json/stringify"));

var _keys = _interopRequireDefault(require("@babel/runtime-corejs2/core-js/object/keys"));

var _path = _interopRequireDefault(require("path"));

var _fs = _interopRequireDefault(require("mz/fs"));

var _latching = _interopRequireDefault(require("latching"));

var _ducky = _interopRequireDefault(require("ducky"));

var _installedPackages = _interopRequireDefault(require("installed-packages"));

var _micromatch = _interopRequireDefault(require("micromatch"));

var _chalk = _interopRequireDefault(require("chalk"));

var _stripAnsi = _interopRequireDefault(require("strip-ansi"));

var _cliTable = _interopRequireDefault(require("cli-table3"));

var _windowSize = _interopRequireDefault(require("window-size"));

var _requireRelative = _interopRequireDefault(require("require-relative"));

/*
**  GemstoneJS -- Gemstone JavaScript Technology Stack
**  Copyright (c) 2016-2019 Gemstone Project <http://gemstonejs.com>
**  Licensed under Apache License 2.0 <https://spdx.org/licenses/Apache-2.0>
*/
const commandHelp = {
  name: "help",
  desc: "Display details about available Gemstone commands",
  opts: [],
  args: [],
  func: async function ()
  /* opts, ...args */
  {
    const pjson = require(_path.default.join(__dirname, "..", "package.json"));

    let out = _chalk.default.bold(`Gemstone Tool ${pjson.version}\n`);

    out += "Copyright (c) 2016-2019 Ralf S. Engelschall <rse@engelschall.com>\n";
    out += "Licensed under Apache License 2.0 <https://spdx.org/licenses/Apache-2.0>\n";
    out += "\n";
    out += `Requested plugin packages: ${_chalk.default.blue(this.packagesUsed.join(", "))}\n`;
    out += "Loaded    plugin packages: ";
    if (this.packagesLoaded.length === 0) out += `${_chalk.default.grey("none")}\n`;else out += `${_chalk.default.blue(this.packagesLoaded.join(", "))}\n`;
    out += "\n";

    const widthsSpread = widths => {
      const widthMax = _windowSize.default.width - 1;
      let total = 0;

      for (let i = 0; i < widths.length; i++) {
        widths[i] = Math.floor(widthMax / 100 * widths[i]) - 1;
        total += widths[i] + 1;
      }

      if (total < widthMax) widths[widths.length - 1] += widthMax - total;
      return widths;
    };

    (0, _keys.default)(this.commands).sort().forEach(cmd => {
      const hasOpts = (0, _keys.default)(this.commands[cmd].opts).length > 0;
      const hasArgs = this.commands[cmd].args.length > 0;
      let table = new _cliTable.default({
        head: [_chalk.default.reset.bold("COMMAND"), _chalk.default.reset.bold("DESCRIPTION")],
        colWidths: widthsSpread([20, 80]),
        style: {
          "padding-left": 1,
          "padding-right": 1,
          border: ["grey"],
          compact: true
        },
        chars: {
          "top-left": "┌",
          "top": "─",
          "top-mid": "┬",
          "top-right": "┐",
          "left": "│",
          "left-mid": "",
          "middle": "│",
          "mid-mid": "",
          "mid": "",
          "right": "│",
          "right-mid": "",
          "bottom-left": hasOpts || hasArgs ? "├" : "└",
          "bottom": "─",
          "bottom-mid": "┴",
          "bottom-right": hasOpts || hasArgs ? "┤" : "┘"
        }
      });
      table.push([_chalk.default.green(cmd), _chalk.default.green(this.commands[cmd].desc)]);
      out += table.toString() + "\n";

      if ((0, _keys.default)(this.commands[cmd].opts).length > 0) {
        table = new _cliTable.default({
          head: [_chalk.default.reset.bold("OPTION"), _chalk.default.reset.bold("TYPE"), _chalk.default.reset.bold("DEFAULT"), _chalk.default.reset.bold("DESCRIPTION")],
          colWidths: widthsSpread([20, 20, 20, 40]),
          style: {
            "padding-left": 1,
            "padding-right": 1,
            border: ["grey"],
            compact: true
          },
          chars: {
            "top-left": "",
            "top": "",
            "top-mid": "",
            "top-right": "",
            "left": "│",
            "left-mid": "",
            "middle": "│",
            "mid-mid": "",
            "mid": "",
            "right": "│",
            "right-mid": "",
            "bottom-left": hasArgs ? "├" : "└",
            "bottom": "─",
            "bottom-mid": "┴",
            "bottom-right": hasArgs ? "┤" : "┘"
          }
        });
        (0, _keys.default)(this.commands[cmd].opts).forEach(name => {
          const opt = this.commands[cmd].opts[name];
          table.push([_chalk.default.blue(opt.name), opt.type, (0, _stringify.default)(opt.def), opt.desc]);
        });
        out += table.toString() + "\n";
      }

      if (this.commands[cmd].args.length > 0) {
        table = new _cliTable.default({
          head: [_chalk.default.reset.bold("ARGUMENT"), _chalk.default.reset.bold("TYPE"), _chalk.default.reset.bold("DESCRIPTION")],
          colWidths: widthsSpread([20, 39, 41]),
          style: {
            "padding-left": 1,
            "padding-right": 1,
            border: ["grey"],
            compact: true
          },
          chars: {
            "top-left": "",
            "top": "",
            "top-mid": "",
            "top-right": "",
            "left": "│",
            "left-mid": "",
            "middle": "│",
            "mid-mid": "",
            "mid": "",
            "right": "│",
            "right-mid": "",
            "bottom-left": "└",
            "bottom": "─",
            "bottom-mid": "┴",
            "bottom-right": "┘"
          }
        });
        this.commands[cmd].args.forEach(arg => {
          table.push([_chalk.default.blue(arg.name), arg.type, arg.desc]);
        });
        out += table.toString() + "\n";
      }
    });
    return out;
  }
};

class Gemstone extends _latching.default {
  constructor({
    verbose = false,
    colored = true
  }) {
    super();
    this.verbose = verbose;
    this.colored = colored;
    this.packagesInstalled = null;
    this.packagesColocated = null;
    this.packagesLoaded = null;
    this.packagesUsed = [];
    this.commands = {};
    this.register(commandHelp);
    this.use("gemstone-tool-plugin-*!");
    this.use("gemstone-*-tool-plugin!");
    this.use("@gemstone-tool-plugin");
  }

  use(...args) {
    args.forEach(arg => {
      if (!(typeof arg === "object" && arg instanceof Array)) arg = [arg];
      const errors = [];
      if (!_ducky.default.validate(arg, "[ string* ]", errors)) throw new Error(`gemstone: ERROR: use: invalid argument(s): ${errors.join("; ")}`);
      this.packagesUsed = this.packagesUsed.concat(arg);
    });
    return this;
  }

  register(spec) {
    this.commands[spec.name] = spec;
    return this;
  }

  async exec(name, opts, ...args) {
    /*  deferred determination of installed packages  */
    if (this.packagesInstalled === null) this.packagesInstalled = await (0, _installedPackages.default)();
    /*  deferred determination of colocated packages  */

    if (this.packagesColocated === null) {
      this.packagesColocated = [];
      let stubLoc;

      try {
        stubLoc = require.resolve(_path.default.join("gemstone", "package.json"));
      } catch (ex) {
        /* NO-OP */
      }

      if (stubLoc) {
        stubLoc = _path.default.dirname(stubLoc);

        let dir = _path.default.join(stubLoc, "node_modules");

        const exists = await _fs.default.exists(dir);
        if (!exists) dir = _path.default.resolve(_path.default.join(stubLoc, ".."));
        const items = await _fs.default.readdir(dir);

        for (let i = 0; i < items.length; i++) {
          let stat = await _fs.default.stat(_path.default.join(dir, items[i]));

          if (stat.isDirectory()) {
            stat = await _fs.default.stat(_path.default.join(dir, items[i], "package.json")).catch(() => null);
            if (stat !== null && stat.isFile()) this.packagesColocated.push(_path.default.join(dir, items[i]));
          }
        }
      }
    }
    /*  deferred loading of requested plugins  */


    if (this.packagesLoaded === null) {
      this.packagesLoaded = [];
      this.packagesUsed.forEach(used => {
        const [, keyword, pattern, optional] = used.match(/^(@)?(.+?)(!)?$/);
        let found = false;
        /*  check standard installation locations  */

        let plugins;
        if (keyword) plugins = this.packagesInstalled.filter(name => {
          let pjson;

          try {
            pjson = (0, _requireRelative.default)(_path.default.join(name, "package.json"), process.cwd());
          } catch (ex) {
            /* NO-OP */
          }

          if (pjson === undefined) {
            try {
              pjson = require(_path.default.join(name, "package.json"));
            } catch (ex) {
              /* NO-OP */
            }
          }

          return typeof pjson === "object" && pjson !== null && typeof pjson.keywords === "object" && pjson.keywords instanceof Array && pjson.keywords.indexOf(pattern) >= 0;
        });else plugins = (0, _micromatch.default)(this.packagesInstalled, pattern, {
          nodupes: true
        });

        if (plugins.length > 0) {
          found = true;
          plugins.forEach(plugin => {
            const obj = (0, _requireRelative.default)(plugin, process.cwd());
            obj.call(this, this);
            this.packagesLoaded.push(plugin);
          });
        }
        /*  check Gemstone colocations  */


        if (keyword) plugins = this.packagesColocated.filter(name => {
          let pjson;

          try {
            pjson = require(_path.default.join(name, "package.json"));
          } catch (ex) {
            /* NO-OP */
          }

          return typeof pjson === "object" && pjson !== null && typeof pjson.keywords === "object" && pjson.keywords instanceof Array && pjson.keywords.indexOf(pattern) >= 0;
        });else {
          plugins = this.packagesColocated.filter(name => {
            const basename = _path.default.basename(name);

            return _micromatch.default.isMatch(basename, pattern, {
              nodupes: true
            });
          });
        }

        if (plugins.length > 0) {
          found = true;
          plugins.forEach(plugin => {
            const obj = require(plugin);

            obj.call(this, this);
            this.packagesLoaded.push(plugin);
          });
        }
        /*  final sanity check  */


        if (!found && !optional) throw new Error(`gemstone: ERROR: exec: no plugin package found for "${used}"`);
      });
    }
    /*  sanity check command  */


    let cmd = this.commands[name];

    if (cmd === undefined) {
      /*  alias matching  */
      const found = (0, _keys.default)(this.commands).filter(command => {
        return typeof this.commands[command].alias === "object" && this.commands[command].alias instanceof Array && this.commands[command].alias.indexOf(name) >= 0;
      });

      if (found.length === 1) {
        name = found[0];
        cmd = this.commands[name];
      }
    }

    if (cmd === undefined) {
      /*  fuzzy matching: prefix matching  */
      const commands = (0, _keys.default)(this.commands);
      let found = commands.filter(command => command.startsWith(name));

      if (found.length !== 1) {
        /*  fuzzy matching: segment matching  */
        const nameSegments = name.split(/-/);
        found = commands.filter(command => {
          const commandSegments = command.split(/-/);
          let match = false;

          if (nameSegments.length === commandSegments.length) {
            let i;

            for (i = 0; i < commandSegments.length; i++) if (!commandSegments[i].startsWith(nameSegments[i])) break;

            if (i === commandSegments.length) match = true;
          }

          return match;
        });
      }

      if (found.length === 1) {
        name = found[0];
        cmd = this.commands[name];
      }
    }

    if (cmd === undefined) throw new Error(`no such command "${name}" registered`);
    /*  sanity check options  */

    (0, _keys.default)(opts).forEach(optName => {
      const opt = cmd.opts.find(opt => opt.name === optName);
      if (opt === undefined) throw new Error("gemstone: ERROR: exec: " + `unknown option "${optName}" on command "${name}"`);
      if (opt.type.match(/^\[.*\]$/) && !(typeof opts[optName] === "object" && opts[optName] instanceof Array)) opts[optName] = [opts[optName]];
      const errors = [];
      if (!_ducky.default.validate(opts[optName], opt.type, errors)) throw new Error("gemstone: ERROR: exec: " + `invalid type "${typeof opts[optName]}" for value of option "${optName}" ` + `on command "${name}": ${errors.join("; ")}`);
    });
    cmd.opts.forEach(opt => {
      if (opts[opt.name] === undefined) {
        if (opt.def !== undefined) opts[opt.name] = opt.def;else throw new Error("gemstone: ERROR: exec: " + `missing mandatory option "${opt.name}" on command "${name}"`);
      }
    });
    /*  sanity check arguments  */

    let i;

    for (i = 0; i < args.length; i++) {
      const spec = cmd.args[i];
      if (spec === undefined) throw new Error(`gemstone: ERROR: exec: too many arguments on command "${name}"`);

      if (spec.type.match(/^\[.*\]$/)) {
        const errors = [];
        if (!_ducky.default.validate(args.slice(i), spec.type, errors)) throw new Error("gemstone: ERROR: exec: " + `invalid arguments #${i}+ on command "${name}": ${errors.join("; ")}`);
        i = args.length;
        break;
      } else {
        const errors = [];
        if (!_ducky.default.validate(args[i], spec.type, errors)) throw new Error("gemstone: ERROR: exec: " + `invalid argument #${i} on command "${name}": ${errors.join("; ")}`);
      }
    }

    const spec = cmd.args[i];
    if (spec !== undefined) throw new Error(`gemstone: ERROR: exec: too less arguments on command "${name}"`);
    /*  execute command  */

    let out = await this.commands[name].func.call(this, opts, ...args);
    if (typeof out !== "string") throw new Error(`gemstone: ERROR: exec: invalid result type on command "${name}" (expected string type)`);
    /*  post-process output  */

    if (!this.colored) out = (0, _stripAnsi.default)(out);
    return out;
  }

}

exports.Gemstone = Gemstone;
module.exports = Gemstone;
